local mod = ...



-- Job Progression: Unwell -------------------------------------------------------------------------------------
mod:register({
    DataType = "JOB_PROGRESSION",
    Id = "JOB_PROGRESSION_PLAGUE_UNWELL",
    ProgressionElementList = {
    }
})
    


-- Job: Unwell -------------------------------------------------------------------------------------
mod:register({
	DataType = "JOB",
	Id = "JOB_UNWELL",
	JobName = "JOB_UNWELL_NAME",
	JobDescription = "JOB_UNWELL_DESC",
    ProductionDelay = 30.0,
    AssetJobProgression = "JOB_PROGRESSION_PLAGUE_UNWELL",
    Hidden = true,
	CharacterSetup = {
		DataType = "CHARACTER_SETUP",
		RightHandObject = {
			DataType = "HAND_OBJECT",
			Model = "PREFAB_TOOL_BARTENDER_WINE_BOTTLE"
		},
		IdleAnimation = "IDLE"
	}
})

mod:override({
    Id = "NEWCOMER",
    CompatibleJobList = {
        Action = "APPEND",
        "JOB_UNWELL"
    }
})

mod:override({
    Id = "SERF",
    CompatibleJobList = {
        Action = "APPEND",
        "JOB_UNWELL"
    }
})

mod:override({
    Id = "COMMONER",
    CompatibleJobList = {
        Action = "APPEND",
        "JOB_UNWELL"
    }
})

mod:override({
    Id = "CITIZEN",
    CompatibleJobList = {
        Action = "APPEND",
        "JOB_UNWELL"
    }
})



-- Building Function: Recover -------------------------------------------------------------------------------------
mod:register({
    DataType = "BUILDING_FUNCTION_WORKPLACE",
    Id = "BUILDING_FUNCTION_PLAGUE_RECOVER",
    Name = "BUILDING_FUNCTION_PLAGUE_RECOVER_NAME",
    WorkerCapacity = 10,
    RelatedJob = { 
        Job = "JOB_UNWELL", 
        Behavior = "BEHAVIOR_WORK"
    },
    InputInventoryCapacity = {
		-- { Resource = "RESOURCE_DISEASE", Quantity = 1 }, 
	},
    ResourceListNeeded = {
		-- { Resource = "RESOURCE_DISEASE", Quantity = 1 }, 
	},
	ResourceProduced = {
        -- { Resource = "RESOURCE_DISEASE", Quantity = 1 },
    }
})



-- Custom Component: Controller -------------------------------------------------------------------------------------
local COMP_PLAGUE = {
	TypeName = "COMP_PLAGUE",
	ParentType = "COMPONENT",
	Properties = {
        { Name = "CurrentDay", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
    }
}



-- Custom Component: for each Villager -------------------------------------------------------------------------------------
local COMP_PLAGUE_VILLAGER = {
	TypeName = "COMP_PLAGUE_VILLAGER",
	ParentType = "COMPONENT",
	Properties = {
        { Name = "DayCreated", Type = "integer", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "FlagUnwell", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
        { Name = "DayUnwell", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
        { Name = "DayRecovered", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
        { Name = "DayImmunityEnd", Type = "integer", Default = 1, Flags = { "SAVE_GAME" } },
        { Name = "JobLast", Type = "JOB", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "WorkplaceLast", Type = "COMP_WORKPLACE", Default = nil, Flags = { "SAVE_GAME" } },
        { Name = "FlagTreated", Type = "boolean", Default = false, Flags = { "SAVE_GAME" } },
    }
}



-- Custom Component: for each Workplace -------------------------------------------------------------------------------------
local COMP_PLAGUE_WORKPLACE = {
	TypeName = "COMP_PLAGUE_WORKPLACE",
	ParentType = "COMPONENT",
	Properties = {
        { Name = "MaxWorkers", Type = "integer", Default = nil, Flags = { "SAVE_GAME" } },
    }
}



-- Custom Component: Villager: Initialise -------------------------------------------------------------------------------------
function COMP_PLAGUE_VILLAGER:init()
end



-- Custom Component: Workplace: Initialise -------------------------------------------------------------------------------------
function COMP_PLAGUE_WORKPLACE:init()
end



-- Custom Component: Controller: Initialise -------------------------------------------------------------------------------------
function COMP_PLAGUE:init()

    mod:log("COMP_PLAGUE ______________________________________________________________________________________")


    local compMainGameLoop = self:getLevel():find("COMP_MAIN_GAME_LOOP")


    -- Daily tasks
    event.register(self, compMainGameLoop.ON_NEW_DAY, function()

        mod:log("Plague: running daily script")

        -- Keep track of the current day
        self.CurrentDay = self.CurrentDay + 1

        -- Add workplace component to track the maximum number of workers ever employed at each place, for use when re-employing recovered workers
        -- mod:log("Plague: adding workplace components")
        self:getLevel():getComponentManager("COMP_WORKPLACE"):getAllComponent():forEach(function(compworkplace)
            compworkplace:getOwner():getOrCreateComponent("COMP_PLAGUE_WORKPLACE")
        end)

        -- Attach villager components to catch any new villagers
        _villagercount = 0
        _physiciancount = 0
        _apothecarycount = 0
        _herbalistcount = 0
        _unwellcount = 0

        self:getLevel():getComponentManager("COMP_VILLAGER"):getAllComponent():forEach(function(compvillager)

            -- Only attach component if the villager has arrived. Avoids a crash where immigrants with no status yet have a component attached
            if (compvillager:isPreInitialized() and compvillager:isInitialized() and compvillager:isEnabled() and compvillager:hasEnabledFlag()) then
                
                _jobcheck = tostring(compvillager:getAssetJob())

                if (_jobcheck ~= "Asset [\"Bailiff.data - JOB\"]" and _jobcheck ~= "Asset [\"Soldier.data - JOB\"]") then -- skip soldiers, monks
                    comphealthvillager = compvillager:getOwner():getOrCreateComponent("COMP_PLAGUE_VILLAGER")

                    -- Record the day that this villager's component was created
                    local compdaycreated = comphealthvillager.DayCreated
                    if (not compdaycreated) then
                        comphealthvillager.DayCreated = self.CurrentDay
                    end

                    -- count physicians and apothecaries for later use
                    if (_jobcheck == "Asset [\"JOB_PHYSICIAN.data - JOB\"]") then
                        _physiciancount = _physiciancount + 1
                    elseif (_jobcheck == "Asset [\"JOB_APOTHECARY.data - JOB\"]") then
                        _apothecarycount = _apothecarycount + 1
                    elseif (_jobcheck == "Asset [\"Herbalist.data - JOB\"]") then
                        _herbalistcount = _herbalistcount + 1
                    end
                    -- Give unwell people the Unwell job
                    if (comphealthvillager.FlagUnwell) then
                        _unwellcount = _unwellcount + 1
                        compvillager:changeJob("JOB_UNWELL")
                    end

                    -- mod:log("Health: Villager " .. _villagercount+1 .. 
                    --     ": FlagUnwell=" .. tostring(comphealthvillager.FlagUnwell) .. 
                    --     ", DayUnwell=" .. tostring(comphealthvillager.DayUnwell) .. 
                    --     ", DayRecovered=" .. tostring(comphealthvillager.DayRecovered) .. 
                    --     ", DayImmunityEnd=" .. tostring(comphealthvillager.DayImmunityEnd) .. 
                    --     ", JobLast=" .. tostring(comphealthvillager.JobLast) ..
                    --     ", WorkplaceLast=" .. tostring(comphealthvillager.WorkplaceLast) ..
                    --     ", FlagTreated=" .. tostring(comphealthvillager.FlagTreated)
                    -- )
                    _villagercount = _villagercount + 1

                end
            end
        end)

        -- Determine Physician ratio to Unwell, to reduce treated patients' recovery times
        if (_physiciancount >= _unwellcount) then
            _physicianratio = 0
        else
            _physicianratio = (1 - (_physiciancount / _unwellcount))
        end

        if (_physicianratio < 0.50) then
            _physicianratio = 0.50
        end

        -- mod:log("Physician count: " .. tostring(_physiciancount))
        -- mod:log("Unwell count: " .. tostring(_unwellcount))
        -- mod:log("Physician ratio: " .. tostring(_physicianratio))

        -- Heal anyone who has come to their Recovered date, and reduce recovery times if attending hospital
        -- mod:log("Plague: heal anyone who has come to their recovered date")
        _currentvillager = 1
        self:getLevel():getComponentManager("COMP_PLAGUE_VILLAGER"):getAllComponent():forEach(
            function(comphealthvillager)

                if (comphealthvillager.FlagUnwell) then 
                    
                    if (comphealthvillager.DayRecovered <= self.CurrentDay) then -- recovery date has come

                        comphealthvillager.FlagUnwell = false
                        comphealthvillager.FlagTreated = false

                        -- Give their old job back
                        if (comphealthvillager.JobLast == nil or not comphealthvillager.JobLast:isValid()) then -- they were unemployed, make them so again
                            comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):changeJob(nil)
                        else

                            local compworkplacecheck = self:getLevel():find(comphealthvillager.WorkplaceLast:getOwner():getId(), "COMP_WORKPLACE")
                            local maxworkers = compworkplacecheck:getOwner():getComponent("COMP_PLAGUE_WORKPLACE").MaxWorkers

                            if (not maxworkers) then
                                maxworkers = 0
                            end

                            if (compworkplacecheck) then -- Check that the workplace still exists
                                -- mod:log("Villager: " .. comphealthvillager:getOwner().Name)
                                -- mod:log("Workplace: " .. compworkplacecheck:getOwner().Name)
                                -- mod:log("Max workers: " .. tostring(maxworkers))
                                -- mod:log("Current worker list: " .. tostring(compworkplacecheck.AssignedWorkerList))

                                -- Check that current worker list is not higher than the max workers ever recorded
                                local workerplace = compworkplacecheck.AssignedWorkerList
                                if (#workerplace < maxworkers) then -- assign them both their old job, and their former workplace as there are less workers now than there have been at some point
                                    comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):changeJob(comphealthvillager.JobLast, comphealthvillager.WorkplaceLast)
                                    -- mod:log("Job changed and workplace assigned.")
                                else -- only assign their old job, but not their old workplace; they will choose one nearby if available
                                    comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):changeJob(comphealthvillager.JobLast)
                                    -- mod:log("Job changed but no workplace assigned.")
                                end
 
                            end
                        end
                        
                        -- mod:log("Villager number " .. tostring(_currentvillager) .. " has recovered from illness.")

                    elseif (not comphealthvillager.FlagTreated) then -- they are sick but not yet treated at hospital

                        -- mod:log(tostring(comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getJobInstance().Workplace:getOwner().Name))
                        -- If they are at hospital, treat them by reducing the recovery time
                        local jobinst = comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getJobInstance()
                        if (jobinst) then -- skip if not assigned to a workplace, it will cause an error that skips this section of code
                            local jobwork = jobinst.Workplace
                            if (jobwork) then
                                if (tostring(comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getJobInstance().Workplace:getOwner().Name) == "BedsPart") then
                                    comphealthvillager.FlagTreated = true
                                    comphealthvillager.DayRecovered = self.CurrentDay + math.floor( ((comphealthvillager.DayRecovered - self.CurrentDay) * _physicianratio) + 0.5 )
                                        -- Find the hospital they are at
                                        -- local compwork = comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getJobInstance().Workplace
                                        -- if (compwork:getOwner():getParent():findFirstObjectWithComponentDown("COMP_RESOURCE_CONTAINER")) then
                                        --     local comptreatment = compwork:getOwner():getParent():findFirstObjectWithComponentDown("COMP_RESOURCE_CONTAINER")
                                        --     if (comptreatment.AvailableQuantity < 100) then -- check for available treatments
                                        --         comptreatment.AvailableQuantity = comptreatment.AvailableQuantity + 1
                                        -- mod:log("Treatment available, number remaining: " .. tostring(100 - comptreatment.AvailableQuantity))
                                        -- mod:log("Current DaysRecovered: " .. tostring(comphealthvillager.DayRecovered))
                                        -- mod:log("Days until DaysRecovered: " .. tostring(comphealthvillager.DayRecovered - self.CurrentDay))
                                        -- mod:log("New DaysRecovered: " .. tostring(self.CurrentDay + math.floor( ((comphealthvillager.DayRecovered - self.CurrentDay) * _physicianratio) + 0.5 )))
                            -- else
                            --     mod:log("Treatment not available.")
                                end
                            end
                        end
                    end
                end
                _currentvillager = _currentvillager + 1
            end
        )
        

        -- Adjust unwell factor based on the number of Apothecaries, up to a maximum of 5 Apothecaries
        _apothecaryfactor = (1 - (_apothecarycount / 10))
        if (_herbalistcount < 1) then
            _apothecaryfactor = 1
        end
        if (_apothecaryfactor < 0.5) then
            _apothecaryfactor = 0.5
        end
        -- mod:log("Apothecary factor: " .. _apothecaryfactor .. " - Apothecaries: " .. _apothecarycount .. " - Herbalists - " .. _herbalistcount)
        
        -- Probability that any one villager will fall ill: between zero and 1% per month, reduced if you have apothecaries
        _unwellfactor = math.random() * (0.01/28) * _apothecaryfactor

        -- Roll dice for each person, if they fall below the unwellfactor then mark them unwell
        -- mod:log("Plague: introduce new incidents of disease")
        if (_villagercount >= 60) then -- Don't get anyone unwell if the population is below 60 (Hospital unlocks at 50)

            self:getLevel():getComponentManager("COMP_PLAGUE_VILLAGER"):getAllComponent():forEach(function(comphealthvillager)

                -- Infect nearby villagers
                if (comphealthvillager.FlagUnwell) then

                    if (comphealthvillager.DayUnwell <= (self.CurrentDay - 2)) then -- only infect after they have been ill for two days

                        local pos1 = comphealthvillager:getOwner():getGlobalPosition() -- position of the sick villager

                        -- Loop through all villagers and look for nearby ones
                        self:getLevel():getComponentManager("COMP_VILLAGER"):getAllComponent():forEach(
                            function(compnearbyvillager)

                                _jobcheck = tostring(compnearbyvillager:getAssetJob())

                                if (_jobcheck ~= "Asset [\"Bailiff.data - JOB\"]" and _jobcheck ~= "Asset [\"Soldier.data - JOB\"]") then -- skip soldiers, monks
                                                        
                                    local pos2 = compnearbyvillager:getOwner():getGlobalPosition()
                                    local distance = math.sqrt( (pos1.x - pos2.x)^2 + (pos1.z - pos2.z)^2 )

                                    if (distance <= 2) then -- villager is less than xm from the unwell person
                                        if (math.random() <= 0.20) then -- x% chance of that person falling ill
                                            makeUnwell(compnearbyvillager:getOwner():getComponent("COMP_PLAGUE_VILLAGER"), self.CurrentDay)
                                            -- mod:log("A nearby villager has been infected.")
                                        end
                                    end

                                end

                            end
                        )

                    end

                end

                if (math.random() <= _unwellfactor) then -- a villager who will become ill

                    _jobcheck = tostring(comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getAssetJob())

                    if (_jobcheck ~= "Asset [\"Bailiff.data - JOB\"]" and _jobcheck ~= "Asset [\"Soldier.data - JOB\"]") then -- skip soldiers, monks
                                            
                        makeUnwell(comphealthvillager, self.CurrentDay)
                    -- mod:log("Villager number " .. tostring(_currentvillager) .. " has fallen ill.")

                    end

                end -- end villager randomly falls ill

            end) -- end loop all villager components

        end -- end pop >= 60

    end) -- end on new day

end -- end init()



-- Function: Set villager unwell -------------------------------------------------------------------------------------
function makeUnwell(comphealthvillager, currentday)

    if (comphealthvillager.DayImmunityEnd <= currentday) then -- do not make them unwell if they are within their immunity period

        if (comphealthvillager.DayCreated <= ( currentday - 5 )) then -- do not make them unwell if they appeared in the last 5 days

            -- mod:log("Plague: make unwell function")

            -- Pick random recovery and immunity times
            _recovery = math.floor(math.random(10, 20) + 0.5)
            _immunity = _recovery + math.floor((_recovery + math.random(20, 200)) + 0.5)

            -- Set villager's health data
            comphealthvillager.FlagUnwell = true
            comphealthvillager.FlagTreated = false
            comphealthvillager.DayUnwell = currentday
            comphealthvillager.DayRecovered = currentday + _recovery
            comphealthvillager.DayImmunityEnd = currentday + _recovery + _immunity

            -- Record the villager's current job
            local currentjob = comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getAssetJob()

            if (not currentjob) then
                _jobId = "nil"
            else
                _job = tostring(currentjob)
                _jobstart = string.find(_job, "\"")
                _jobend = string.find(_job, ".data")
                _jobId = string.gsub(string.upper(string.sub(_job, _jobstart + 1, _jobend - 1)), " ", "_") -- extract the JOB Id from the string that comes from the component
            end

            if (_jobId ~= "JOB_UNWELL") then -- Don't record the job if they are currently sick

                -- Record the job
                comphealthvillager.JobLast = currentjob

                -- Record the workplace
                if (currentjob) then
                    local jobinst = comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):getJobInstance()
                    if (jobinst) then
                        local jobwork = jobinst.Workplace
                        if (jobwork) then

                            -- Record the current workplace component
                            comphealthvillager.WorkplaceLast = jobinst.Workplace

                            -- Count the current workers
                            local workerlist = jobinst.Workplace.AssignedWorkerList

                            -- Record the max number of workers, if it is higher than the current value
                            local maxworkers = jobinst.Workplace:getOwner():getComponent("COMP_PLAGUE_WORKPLACE").MaxWorkers
                            if (maxworkers) then
                                if (maxworkers <#workerlist) then
                                    jobinst.Workplace:getOwner():getComponent("COMP_PLAGUE_WORKPLACE").MaxWorkers = #workerlist
                                end
                            else
                                jobinst.Workplace:getOwner():getComponent("COMP_PLAGUE_WORKPLACE").MaxWorkers = #workerlist
                            end

                        end
                    end
                end

                -- Change job to Unwell
                -- I believe this was causing an infrequent crash when an Unwell job was being assigned to a villager that was immigrating but hadn't arrived yet
                -- I got an error that would crash: PlaCDataVillager::initCurrentStatus        Villager has no reached status
                -- By removing the job assignment to Unwell, it makes a space until the next day when the immigrant will presumably have a status and can be assigned Unwell
                -- The next day after they are marked unwell in the component, they will be marked with the Unwell job
                -- comphealthvillager:getOwner():getComponent("COMP_VILLAGER"):changeJob("JOB_UNWELL")

            end
        end
    end
end



-- Register component class
mod:registerClass(COMP_PLAGUE)
mod:registerClass(COMP_PLAGUE_VILLAGER)
mod:registerClass(COMP_PLAGUE_WORKPLACE)



-- Attach component to the prefab manager (a prefab that comes built into the game, for this purpose)
mod:registerPrefabComponent("PREFAB_MANAGER", { DataType = "COMP_PLAGUE" } )
